(* File:    counterAutomata.sml
   Content: - functor ppCFm extends a (deterministic) finite automaton M with a counter
              holding the set of positions in the input text w where a prefix v of w with 
              v in L(M) ends.
            - functor Matcher turns a regular expression into a (minmal deterministic) 
              finite automaton with counter that reports the (end position of) matches 
              found in a text.
   Author:  Hans Leiss, CIS, leiss@cis.uni-muenchne.de
   Date:    16.12.99, cleaned up 28.12.99
   Edited:  24.02.00  separated functor Matcher to pattern-matcher.sml
            05.05.08  changed "ghostview --magstep -2 " to "gv --scale=-2 "

   Changes to ppfm.sml:
   - `abstype machine' includes fields 
           position : int ref,   hits : int list ref
   - `position' stores the current position in the input list
   - `hits' tells the positons in the input where M was in a final state
   - `nextstates' increments the position and hits counters

   This file is used in kmp-automaton.sml to define a Knuth-Morris-Pratt pattern matcher
   for single pattern, and in matcher.sml to define search machine for regular pattern search.
*)

signature PPCFM =
 sig
  eqtype label 

  type ''a machine                                   
  val accept : ''a machine -> ''a list -> bool 
  val hits : ''a machine -> ''a list -> int list  (* accepted prefixes *)
  val make : {trans : (int * ''a * int) list, 
              start : int list, 
              final : int list} -> ''a machine
  val show : ''a machine -> {start : int list, 
                         trans : (int * ''a * int) list, 
                         final : int list}
  val display : label machine -> string -> OS.Process.status
 end;

functor ppCFm(Alphabet : LABEL) :> PPCFM where type label = Alphabet.label =
  struct
    local 
      fun member x l = List.exists (fn y => y = x) l
      and add q [] = [q]
        | add q (u::U) = if q=u then (u::U) else (u::(add q U))
      and union [] U = U
        | union (q::Q) U = union Q (add q U)
      and intersects [] Q = false
        | intersects (p::P) Q = if member p Q then true 
                                else intersects P Q
      and Union [] = []
        | Union (Q::Qs) = union Q (Union Qs)

    in         
      type label = Alphabet.label 

      abstype 'a machine = 
        Ma of {states : int,
               start : int Set.t, 
               trans : (int * 'a * int) Set.t, 
               final : int Set.t,
               position : int ref,
               hits : int list ref}
      with

        fun make {trans=I, start=S, final=F} =
          let 
            val maxStart = foldr Int.max 0 S
            val maxFinal = foldr Int.max maxStart F
            val maxState = 
              foldr (fn ((i,a,j),k) => 
                     Int.max (Int.max(i,j),k)) maxFinal I
          in 
            Ma{states = maxState+1, start = Set.mk S, trans = Set.mk I, final = Set.mk F,
               position = ref 0, hits = ref ([] : int list)}
          end

        fun show (M as Ma{position = p, hits = h, ...}) = 
            {start = start M, trans = trans M, final = final M} (* , position = !p, hits = !h} *)
        and start  (Ma{start = S, ...}) = Set.list S
        and final  (Ma{final = F, ...}) = Set.list F
        and trans  (Ma{trans = I, ...}) = Set.list I
        and states (Ma{states = n, ...}) = n
        and incrPos (M as Ma{position = p, ... }) = (p := !p +1; M)
        and addHit (Ma{position = p, hits = h, ...}) = (h := (!p)::(!h))

        fun nextstates M q a = 
            let fun g [] acc = acc
                  | g ((x,b,p)::r) acc =
                if (x,b) = (q,a) then g r (add p acc) else g r acc
                val reached = g (trans M) []
            in 
                incrPos M;
                if intersects reached (final M) then addHit M else ();
                reached
            end  
        and reach M Q [] = Q
          | reach M Q (a::l) = 
            let val Qs = map (fn q => (nextstates M q a)) Q 
            in reach M (Union Qs) l end
        and accept M w = intersects (reach M (start M) w) (final M)

        and hits (M as Ma{hits = h,...}) w = (accept M w; !h)

        and labels M = 
            let fun labls ((x,a,y)::Trans) L = labls Trans (add a L)
                  | labls [] L = L
            in labls (trans M) [] end

        fun draw M os name =
            let
                fun write w = TextIO.output (os, w)
                fun drawTransition (i,a,j) =
                    write (Int.toString i ^ " -> " ^ Int.toString j 
                           ^ " [label = " ^ (Alphabet.toString a) ^ "];\n")
                fun drawInitialState i = 
                    write (Int.toString i ^ " [shape = diamond] ;\n")
                fun drawFinalState i =   
                    if (List.exists (fn x => x=i) (start M)) 
                        then write (Int.toString i ^ " [shape = diamond, peripheries = 2];\n")
                    else write (Int.toString i ^ " [shape = doublecircle];\n")
            in 
                write ("digraph " ^ name ^ 
                       " {center=true;\n size=\"7.5,14\";\n ratio=compress;\n");
                app drawTransition (trans M);
                app drawInitialState (start M);
                app drawFinalState (final M);
                write "}\n\n"
            end

        fun dot M filename =
            let
                val os = TextIO.openOut (filename ^ ".tmp.dot")
            in
                draw M os filename;
                TextIO.closeOut os
            end

        fun showgraph filename =
            (OS.Process.system 
             ("dot -Tps " ^ filename ^ ".tmp.dot -o " ^ filename ^ ".tmp.ps ");
             (* OS.Process.system ("ghostview -magstep -2 " ^ filename ^ ".tmp.ps &") *)
             OS.Process.system ("gv --scale=-2 " ^ filename ^ ".tmp.ps &")
            )
      end

      fun display M filename = (dot M filename; showgraph filename)
    end
  end


