(* Normal forms of a regular system { x0 = re0, ..., xm = rem }  H.Leiss, 11/97 
   under the language interpretation (mostly).                   3/99

   [Had been: grammarNorm.sml, regeqnsNF.sml]

   22.04.98 - adapted to simpler definition of regular expressions
   19.02.99 - Pretty printing: now in RegExp.
   18.11.99 - follows k corrected, but still too slow. (for k > 2 at least)
   30.11.99 - Changed some precomputations of first, but this did not improve performance.
              Lost in updating files
    5.12.99 - Compute (fixBool eval env eqs) by starting iteration with 
           
                  env'(i) := if (recVar i eqs) then false else (env i)

              instead of env, because eqs need not be increasing. For example, 
              (fixBool nonEmpty [true,false] [X = XY, Y=Y]) has to produce [false,false]
   14.02.00 - moved RegMod to a separate file, regmod.sml
            - moved Language and KA to a separate file, Develop-Lib/language.sml

   Under Development: 3/99
   - Generalize properly to allow free non-recursion variables.
   - Checked rlc for lr(0)-parsing. Ok if there are no free variables in the grammar.

   To be done:
   - better top-level calls for rlc-computations.
    
Compiler.Control.Print.printDepth := 20;             
*)

signature GRAMMARS = (* EBNF-Grammars and their properties *)
  sig
    type 'a t = 'a list Set.t  (* languages over alphabet 'a *)

    type 'a env = int -> 'a
    type 'a eqns = (int * 'a) list

    datatype re = datatype RegExp.re
    val Unit : 'a re
    val Zero : 'a re

    val maxVar : 'a re eqns -> int
    val recVars : 'a re eqns -> int list
    val freeVars : 'a re eqns -> int list

    val unit : unit -> ''a t   (* val unit : ''a t raises problems with value restriction,
                                  val zero = Set.empty : ''a t  does not !?? *)
    val zero : unit -> ''a t
    val atom : ''a -> ''a t

    val initial : int -> ''a t -> ''a t
    (* 
    val fixBool : (bool env -> 'a re -> bool)
                  -> bool env -> 'a re eqns -> bool env
    val fixLan : (''a t env -> ''a re -> ''a t)
                 -> ''a t env -> ''a re eqns -> ''a t env
    val solveBool : (bool env -> 'a re -> bool)
                    -> bool env -> 'a re eqns -> bool eqns
    val solveLan : ('a t env -> 'a re -> 'a t)
                   -> 'a t env -> 'a re eqns -> 'a t eqns
    *) 
    val geUnit  : bool env -> 'a re eqns -> 'a re -> bool
    val nonEmpty  : bool env -> 'a re eqns -> 'a re -> bool

    (* val subtractUnit  : bool env -> 'a re eqns -> 'a re -> 'a re *)
    val subtractUnits : bool env -> ''a re eqns -> ''a re eqns

    val nfEmpty   : bool env -> ''a re eqns -> ''a re eqns

    (* val Fst       : int -> ''a t env -> ''a re eqns -> ''a t env *)
    val first     : int -> ''a t env -> ''a re eqns -> ''a re -> ''a t
    (* val evalfst   : int -> ''a t env -> ''a re -> ''a t *)

    val leftcontexts  : ''a re eqns -> ''a re -> ''a re eqns

    (* val fol      : int -> ''a re eqns
                       -> ''a re -> ''a t env -> ''a re -> ''a t *)
    val follows  : int -> ''a t env
                       -> ''a re eqns -> ''a re -> ''a re -> ''a t
    (* val follsFix : int -> ''a t env -> ''a re eqns -> ''a re -> ''a t env *)
  end 

structure Grammars : GRAMMARS = 
    struct

	(* From the polymorphic Kleenean Algebra operations, we only need: *)

	type 'a t = 'a list Set.t

	fun zero () = Set.empty  : ''a t  
	and unit () = Set.mk[[]] : ''a t
	and atom a = Set.mk[[a]] : ''a t

	(* interpretation of recursion equation systems *)

        type 'a env = (int -> 'a)
        and  'a eqns = (int * 'a) list

	datatype re = datatype RegExp.re
	val Unit = RegExp.Unit
	and Zero = RegExp.Zero

        val recVars = RegEqns.recVars 
        and freeVars = RegEqns.freeVars
	and maxVar = RegEqns.maxVar

        fun fixBool eval (init : bool env) (eqs : 'a re eqns) =
            RegEqns.fix eval init (fn Env => map Env (recVars eqs)) eqs : bool env
        and solveBool eval (init : bool env) (eqs : 'a re eqns) =
            RegEqns.solve eval init (fn Env => map Env (recVars eqs)) eqs

        fun fixLan eval (init :  ''a t env) (eqs : ''a re eqns) : ''a t env = 
            RegEqns.fix eval init (fn Env => map (Set.size o Env) (recVars eqs)) eqs 
        fun solveLan eval (init : 'a t env) (eqs : 'a re eqns) = 
            RegEqns.solve eval init (fn Env => map (Set.size o Env) (recVars eqs)) eqs

        (* -------- Properties of r resp. L(G)(r) for G={Xs=rs} -------------- *)

        fun geUnit   env eqs r = geUn (fixBool geUn (initEnv env eqs) eqs) r
        and geUnits  env eqs   = map (fn (i,r) => (i,geUnit env eqs r)) eqs
        and nonEmpty env eqs r = nonEm (fixBool nonEm (initEnv env eqs) eqs) r
        and nonEmptys env eqs  = map (fn (i,r) => (i,nonEmpty env eqs r)) eqs
            (* Compute (fixBool eval env eqs) by starting iteration with 
               (initEnv env eqs) instead of env, because eqs need not be increasing. 
               So (fixBool nonEmpty [true,false] [X = XY, Y=Y]) gives [false,false] *)
        and initEnv env eqs i = 
            if List.exists (fn k => k=i) (recVars eqs) then false else env i

        and geUn env (Atom a) = false
          | geUn env (Var i) = env i 
          | geUn env (Plus []) = false
          | geUn env (Plus (r::rs)) = (geUn env r) orelse (geUn env (Plus rs))
          | geUn env (Times []) = true
          | geUn env (Times (r::rs)) = (geUn env r) andalso (geUn env (Times rs))
          | geUn env (Star r) = true

        and nonEm env (Atom a) = true
          | nonEm env (Var i) = env i
          | nonEm env (Plus []) = false
          | nonEm env (Plus (r::rs)) = (nonEm env r) orelse (nonEm env (Plus rs))
          | nonEm env (Times []) = true
          | nonEm env (Times (r::rs)) = (nonEm env r) andalso (nonEm env (Times rs))
          | nonEm env (Star r) = true


        (* ----- Compute an expression (r-1) relative to system Xs=rs ---------

         Replace r(X) by (r-1)(X') where for recursion variable X of Xs:
         X' = X   if env(X) = 0 (X did not contain 1) and
         X' = X-1 if env(X) = 1 (X contained 1) in the solution of Xs=rs

         This is also done for free variables, where one probably has to 
         *demand* that env(X) = 0!  Why not use falseEnv?? *)

        fun subtractUnits env eqs : ''a re eqns =  
            map (fn (i,r) => (i,subtractUnit env eqs r)) eqs 

        and subtractUnit (env : bool env) (eqs : ''a re eqns) (e:''a re) =
            RegExp.nf (subUn (fixBool geUn env eqs) e)  

        and subUn env (Atom x) = (Atom x)
          | subUn env (Var i) = (Var i)  (* ! ok for recursion vars, but else? *)
          | subUn env (Plus rs) = Plus (map (subUn env) rs)
          | subUn env (Times []) = Zero
          | subUn env (Times (r::rs)) =
            (* 
               Because the meaning of X has to change to X-1 everywhere, all 
               factors r have to be replaced by (r-1) resp. (r-1)+1 and the 
               results must be summed up. 
             *)
            if geUn env r 
                then if not(geUn env (Times rs))
                         then Times [Plus [Unit,subUn env r], subUn env (Times rs)]
                     else Plus [Times [Plus [Unit, subUn env r], 
                                       subUn env (Times rs)],
                                Times [subUn env r, 
                                       Plus [Unit, subUn env (Times rs)]]]
            else if (geUn env (Times rs))
                     then Times [subUn env r, Plus [Unit, subUn env (Times rs)]]
                 else Times [subUn env r, subUn env (Times rs)]
          | subUn env (Star r) = Times [subUn env r, Star (subUn env r)]


        (* --- Simplify a system Xs=rs by setting empty components to Empty ---- *)

        fun nfEmpty env eqs :''a re eqns = (* env(i)=true iff L(Var i)=Empty *)
            let 
                fun isEmpty (i,r) = not (nonEmpty (fn j => not (env j)) eqs r)
                val emptyRecVars = map #1 (List.filter isEmpty eqs)
                val emptyFreeVars = 
                    List.filter (fn i => env i = true) (freeVars eqs)

                val emptys = emptyRecVars @ emptyFreeVars
                val eqns = map (fn i => (i,Zero)) emptys
                fun subst (i,r) = 
                    case (List.find (fn j => j=i) emptys) of
                        SOME j => (i,Zero)
                      | NONE => (i,RegExp.nf (RegExp.subst eqns r))
            in 
                map subst eqs
            end        

        (* --------------- Auxiliary computations for LL(k)-parsers -------------- *)

	(* --- first k symbols of words in language [[r]]env ---    *)

	fun evalfst k (env : ''a t env) (Var i) = initial k (env i) 
	  | evalfst k env (Atom a) = if k>0 then (atom a) else zero()
	  | evalfst k env (Plus rs) = Set.Union (map (evalfst k env) rs) 
	  | evalfst k env (Times rs) = cat k (map (evalfst k env) rs)
	  | evalfst k env (Star r) = evalfst k env (RegExp.Powers r k)
	and initial k L = 
	    let 
		fun init k [] = zero ()
		  | init k (w::ws) = Set.add (prefix k w) (init k ws)
	    in 
		init k (Set.list L)
	    end 
          
        and prefix k [] = []
          | prefix k (a::aa) = if k > 0 then a::(prefix (k-1) aa) else []
	and concatfst k u v = (* first k elements of (u @ v) *)
	    let 
		val n = length u 
	    in
		if k <= n then prefix k u
		else if length v > k-n then u @ (prefix (k-n) v) else u @ v
	    end
        and cat k (S::Ss) = (* (initial k (concatenate (S::Ss))) *)
            let 
             fun catk (w::ws,L::Ls) acc = if L = [] then [] else
               catk (ws,L::Ls) ((map (concatfst k w) L) @ acc)
               | catk ([],L::Ls) acc = if L = [] then [] else catk (acc,Ls) []
               | catk (L,[]) [] = (map (prefix k) L) 
               | catk (L,[]) _ = L (* does not occur ? *)
            in 
              Set.mk (catk (Set.list S, map Set.list Ss) [])
            end
          | cat k [] = unit ()

	fun first k env (eqs : ''a re eqns) r = evalfst k (Fst k env eqs) r

	and Fst k env eqs = fixLan (evalfst k) (fn i => initial k (env i)) eqs


        (* ----------------------- follow k eqs s r ---------------------------------
 
           For regular system eqs and number k, compute the words of length <= k that
	   follow r-expressions in expression s:

	   follow k eqs s r :=   { k(w) | uvw in L(s), v in L(r) } 
	   
                                 with L(s) = lan(s)[fixLan eqs]

           Let As be approximations of {follow k eqs X_1 r,...,follow k eqs X_m r}.
	   Then (fol k eqs s As r) should be (follow k eqs s r) relative to As.

	   Having that, one can iterate it to get (follow k eqs X_i r) and then 
	   (follow k eqs s r).

        *)   

	fun fol k eqs s (env : ''a t env) r : ''a t = 
	    let 
             fun fst k r = evalfst k (Fst k (fn i => zero()) eqs) r
	    in 
		case s of
		    (Var i) => if r=s then Set.union (env i) (unit()) else (env i)
		  | (Atom a) => if r=s then unit() else zero()
		  | (Plus []) => zero()
		  | (Plus ss) => 	 
			let 
			    val L = Set.Union (map (fn e => fol k eqs e env r) ss) 
			in 
			    if r=s then Set.union (unit()) L else L
			end
		  | (Times []) => if r=s then unit() else zero()
		  | (Times (e::es)) => 
			let val L = 
			    Set.union (cat k [fol k eqs e env r, fst k (Times es)])
			              (fol k eqs (Times es) env r)
			in 
			    if r=s then Set.union (unit()) L else L 
			end
		  | (Star e) => 
			let 
			    val L = fol k eqs (RegExp.Powers e (k+1)) env r
			in 
			    if r=s then Set.union (unit()) L else L
			end
	    end

	and follsFix k env eqs r =
	    fixLan (fn Env => fn s => fol k eqs s Env r) env eqs
	    
	and follows k env eqs s r = 
	    fol k eqs s (follsFix k env eqs r) r


        (* **** Reduced left context of t in recursion variables of grammar eqs *******

           We need new recursion variables recVars' to define the recursion equations 
           for the reduced left contexts of t in s relative to eqs, where for Sig :=
           (Vars+Atoms eqs)

	       rlc of t in s relative to eqs :=
	           { w in Sig* | s =>rm w t beta , for some beta in Sig* }

           The variables of eqs may occur as parameters in the new equations, so we 
           use new variables (Var i+n) (for rlc (Var i) t) with 
           n = 1+ max{recVars eqs, freeVars eqs} as recursion variables of the 
           system 
	       
	       eqs'  =  { rlc (Var i) t = rlc ri t | i in I }.

           These are the recursive definition of the reduced left context of t in 
           the recursion variables of eqs.
	*)

	fun rlc eqs t =              (* assume eqs have recursion variables only *)
	    let                      (* lc s = rlc of t in s *)    
		val n = 1 + (maxVar eqs)

		fun isRecVar (Var k) = 
		    List.exists (fn j => j=k) (recVars eqs)
		  | isRecVar _ = false
		    
		fun lc s = if t = s andalso not(isRecVar t) then Unit 
			   else case s of
			       (Atom a) => Zero
			     | (Var i) => 
				   (if isRecVar s then (Var (i+n)) 
				    else s) (* better: else env(i) for parameter vars *)
			     | (Plus ss) => Plus (map lc ss)
			     | (Times []) => Zero
			     | (Times (s::ss)) => Plus [lc s, Times [s, lc (Times ss)]]
			     | (Star s) => Plus [lc Unit, Times [Star s, lc s]]

                fun newEqn (i,e) = (i+n, if t=(Var i) then Plus[Unit,lc e] else lc e)
	    in
		map newEqn eqs 
	    end

	and leftcontexts eqs t = map (fn (i,e) => (i,RegExp.dnf e)) (rlc eqs t)
    end

(* This is ok for computing the reduced left context of t in the recursion variables 
   of the grammar eqns - if there are no parameter variables. For eqs with free 
   variables, some environment env : (int -> 'a re) with env(j) = (rlc (Var j) t) 
   has to be added.
*)



(*

functor ppGRAMMARS(val varName : int -> string) : PPREGEXP where type alph = string
    = ppREGEXP(type alph = string 
	       fun atomName a = a
	       val varName = varName);

functor eBNF (eqtype alph) =
    struct
	type language = alph list Set.t
	val exp = alph RegExp.re
        val lan : int -> (int -> language) -> exp -> language
    end
*)


