signature kaMATRIX = (* Kleenean matrix = Matrix over Kleenean Algebras *)
    sig
	structure Ring : KA
	type t 
	val dimension : t -> int * int
	val tabulate : (int * int) -> (int * int -> Ring.t) -> t
	val sub : t * (int * int) -> Ring.t
	val map : (Ring.t -> Ring.t) -> t -> t

	val zero : int * int -> t
	val unit : int * int -> t
	val sum : t * t -> t
	val prod : t * t -> t
	val star : t -> t
    end 

functor kaMatrix (R : KA) : kaMATRIX = 
    struct
	structure Ring = R
	type r = Ring.t  

	type t = r vector vector 

	fun dimension (S:t) = 
	    let 
		val rows = Vector.length S
		val columns = Vector.length (Vector.sub(S,0)) 
		val samelengths = 
		    Vector.foldr (fn (v,b) => b andalso (Vector.length v = columns)) 
		                 true S
	    in 
		if samelengths then (rows,columns) else raise Size 
	    end

	fun sub (E:t,(i,j)) = 
	    let 
		val (d1,d2) = dimension E 
	    in 
		if 0 <= i andalso i < d1 andalso 0 <= j andalso j < d2 
		    then Vector.sub (Vector.sub (E,i), j)
		else raise Size
	    end

	fun tabulate (rows,cols) (f : int * int -> r) =
	    Vector.tabulate(rows, fn i => Vector.tabulate (cols, fn k => f(i,k))) : t

	fun map (f : r -> r) (A : t) =
	    let 
		val (n,m) = dimension (A:t)
	    in
		tabulate (n,m) (fn (i,j) => f (sub (A,(i,j))))
	    end

	fun zero (n,m) = tabulate (n,m) (fn (i,k) => R.zero) 
	fun unit (n,m) = tabulate (n,m) (fn (i,k) => if i=k then R.unit else R.zero) 

	fun sum (S:t,T:t) : t = 
	    let
		val (ns,ms) = dimension S
		val (nt,mt) = dimension T
		fun sumV (v:r vector,w:r vector) : r vector = 
		    Vector.tabulate (Vector.length v, 
				     fn i => R.sum (Vector.sub(v,i),
						    Vector.sub(w,i))) 
	    in
		if (ns,ms)=(nt,mt) then
		    Vector.tabulate (ns, fn i => sumV (Vector.sub (S,i),Vector.sub(T,i)))
		else raise Size
	    end

	fun prod (S:t,T:t) : t =
	    let 
		val (ns,ms) = dimension S
		val (nt,mt) = dimension T
		fun prodV  (v:r vector,w:r vector) : r = 
		    let 
			fun sm i = if i=0 then R.zero 
				   else R.sum(sm (i-1), 
					      R.prod (Vector.sub(v,i-1),Vector.sub(w,i-1)))
		    in 
			sm (Vector.length v)
		    end
		fun column (T:t,j) = Vector.map (fn r => Vector.sub (r,j)) T
 	    in
		if ms=nt then
		    tabulate (ns,mt) (fn (i,j) => prodV (Vector.sub (S,i), column(T,j)))
		else raise Size
	    end

	fun star (E:t) : t =
	    let
		val (d,e) = dimension E
	    in 
		if d=e then
		    if d=1 then tabulate (1,1) (fn (i,j) => R.star (sub (E,(i,j))))
		    else
			let
			    val A = tabulate (1,1)     (fn (i,j) => sub (E,(i,j)))
			    val B = tabulate (1,d-1)   (fn (i,j) => sub (E,(i,j+1)))
			    val C = tabulate (d-1,1)   (fn (i,j) => sub (E,(i+1,j)))
			    val D = tabulate (d-1,d-1) (fn (i,j) => sub (E,(i+1,j+1)))
			    val Ds  = star D
			    (* recursion:  E* = [[F*, F*BD*], [D*CF*, D* + D*CF*BD*]  *)
			    val BDs = prod(B,Ds)
			    val F   = sum(A,prod(BDs,C))
			    val A' = star F
			    val B' = prod(A', BDs)
			    val C' = prod(Ds, prod(C,A'))
			    val D' = sum(Ds, prod(C',BDs))
			in
			    tabulate (d,d) (fn (i,j) => 
					    if i < 1
						then if j < 1 then sub (A',(i,j))
						     else sub (B',(i,j-1))
					    else if j < 1 then sub (C',(i-1,j))
						 else sub (D',(i-1,j-1)))
			end
		else raise Size
	    end
    end

(* Example:

structure PathMat = kaMatrix(Path);

fun fast_path M =
    let
	val (n,k) = PathMat.dimension M
	fun f (m,M) = if n-1 <= m then M
			else f(2*m, PathMat.prod(M,M))
    in
	f (1,M) 
    end;

let 
    val z = Path.zero 
    val P = (PathMat.tabulate) (5,5) 
                 (fn (0,0) => 0 | (0,1) => 3 | (0,2) => 8 | (0,3) => z | (0,4) => ~4
	           | (1,0) => z | (1,1) => 0 | (1,2) => z | (1,3) => 1 | (1,4) => 7
		   | (2,0) => z | (2,1) => 4 | (2,2) => 0 | (2,3) => z | (2,4) => z
		   | (3,0) => 2 | (3,1) => z | (3,2) => ~5| (3,3) => 0 | (3,4) => z
		   | (4,0) => z | (4,1) => z | (4,2) => z | (4,3) => 6 | (4,4) => 0
		   | _ => raise Size)
in 
    PathMat.star P
end

(* all_pairs_shortest_paths:  (fast_path P = PathMat.star P)

*)
*)
