(* Knuth-Morris-Pratt-Automaton for pattern p           14.12.99 H.Leiss

   Given a single pattern string p = p1 ... pm : ''a list, we build a list 
   of m+1 states 0,...,m with transitions (i,p(i+1),i+1), the `skeleton' of 
   the KMP-automaton. 

   On finding the prefix p1...pj of the pattern p ending in position k+j 
   of a text t = a1....an, the KMP-automaton reaches state (j+1). 
   If the next symbol a(k+j+1) of t is not p(j+1), KMP goes to failure(j,p), 
   the largest r<j such that a(k+j-r +1)...a(k+j) is a prefix of p, and uses 
   transition (r,a(k+j+1),q).

   The automaton has a counter to report occurrences (by initial position) 
   of a pattern found in a text. (We allow search for a single pattern only).

   Displaying of the automaton uses the graphviz/dot-program of AT&T.

   Bugs: 
   - finding the empty string does not report the occurrence at position 0

   Edited: 
   27.12.99 
   - adapted to set.sml and automata.new.sml 
   - added restriction :> to simplified signatures
   - added Remark on KMP = minimize(determine(retofm(alphabet^*;pattern)))
*) 

(*  See ~/SML/kmp-automata.cm  *)

signature KMP =
  sig
    type t 
    datatype s = Dummy | Pat of t

    structure M : PPCFM where type label = s
    val kmp : t list -> s M.machine
    val display : t list -> OS.Process.status
    val find : (t list) -> (t list) -> int list
  end

signature simpleKMP = (* assuming that all text symbols occur in the pattern *)
  sig
    structure M : PPCFM
    type t
    val display : t list -> OS.Process.status
    val kmp : t list -> t M.machine
    val find : (t list) -> (t list) -> int list
  end

functor KMPautomaton(Alphabet : LABEL) :> KMP where type t = Alphabet.label =
    struct 
        type t = Alphabet.label
        datatype s = Pat of t | Dummy       (* for edges with labels not in pattern *)

        structure NewAlphabet = 
            struct 
                type label = s 
                fun toString Dummy = "_"  (* for .dot-drawings *)
                  | toString (Pat(t)) = Alphabet.toString t
            end

        structure M = ppCFm(NewAlphabet) : PPCFM  (* finite automata over NewAlphabet *)

        fun pats [] acc = acc
          | pats (p::ps) acc = 
            if (List.exists (fn x => x=p) acc) then (pats ps acc)
            else pats ps (p::acc)

        fun kmp(pattern : t list) =
            let 
                val patternsymbols = pats pattern []
                val labels = (Dummy :: (map (fn t => Pat(t)) patternsymbols))

                val positions = List.tabulate (length pattern + 1,fn x => x)

                fun failure(j,pattern) = (* ok *)
                    if j=0 then 0 
                    else 
                        let 
                            val prefix = List.take (pattern,j)
                            fun good s = 
                                if 0 <= s andalso s < j andalso
                                    (List.take (prefix,s) = List.drop (prefix,j-s)) 
                                    then s else 0
                        in
                            foldl Int.max 0 (map good positions)
                        end

                fun nextstate pattern (j,c) =
                    if j < (length pattern) andalso c = List.nth (pattern,j) 
                        then j+1 else
                            if j = 0 then 0 else 
                                nextstate pattern (failure(j,pattern),c) 

                fun mkTrans (j,a) = (* Dummies do not subsume Pat(_) in M *)
                    case a of 
                        Pat(c) => let
                                      val k = nextstate pattern (j,c) 
                                  in
                                      (* if k=0 then NONE else *) SOME (j,Pat(c),k)
                                  end
                      | Dummy =>  SOME (j,Dummy,0)

                fun transits j = 
                    let
                        fun good (SOME x) = true | good NONE = false
                        val (aux,_) = List.partition good (map (fn a => mkTrans(j,a)) labels)
                    in 
                        map (fn (SOME x) => x) aux
                    end
                val transitions = List.concat (map transits positions)      

            in 
                M.make{start=[0],trans=transitions,final=[length pattern]}
            end

        fun display (pattern : t list) = M.display (kmp pattern) "kmp"

        fun find (pattern : t list) (text : t list) =
            let 
                val PM = kmp pattern
                val len = length pattern
                val patternsymbols = pats pattern []
                fun convert x = if List.exists (fn y => y=x) patternsymbols 
                                   then Pat(x) else Dummy
            in
                List.rev (map (fn i => i-len) (M.hits PM (map convert text)))
            end
    end

functor simpleKMPautomaton(Alphabet : LABEL) :> simpleKMP where type t = Alphabet.label =
    (* assumes: input text has no symbols except those in pattern *)
    struct 
        type t = Alphabet.label

        structure M = ppCFm(Alphabet) : PPCFM  (* finite automata over Alphabet *)

        fun kmp(pattern : t list) =
            let 
                fun patsymbols [] acc = acc
                  | patsymbols (p::ps) acc = 
                    if (List.exists (fn x => x=p) acc) then (patsymbols ps acc)
                    else patsymbols ps (p::acc)

                val positions = List.tabulate (length pattern + 1,fn x => x)
                val labels = patsymbols pattern []

                fun failure j = 
                    if j=0 then 0 
                    else 
                        let 
                            val prefix = List.take (pattern,j)
                            fun good s = 
                                if 0 <= s andalso s < j andalso
                                    (List.take (prefix,s) = List.drop (prefix,j-s)) 
                                    then s else 0
                        in
                            foldl Int.max 0 (map good positions)
                        end

                fun nextstate (j,c) =
                    if j < (length pattern) andalso c = List.nth (pattern,j) 
                        then j+1 else
                            if j = 0 then 0 else 
                                nextstate (failure j,c) 

                fun transits j = map (fn a => (j,a,nextstate(j,a))) labels
                val transitions = List.concat (map transits positions)      

            in 
                M.make{start=[0],trans=transitions,final=[length pattern]}
            end

        fun display (pattern : t list) = M.display (kmp pattern) "kmps"

        fun find (pattern : t list) (text : t list) =
            let 
                val PM = kmp pattern
                val len = (length pattern)
            in
                List.rev (map (fn i => i-len) (M.hits PM text))
            end
    end


(* Example 1: full KMP-automaton:

local 
    structure KMP = KMPautomaton(CharLabels)
in 
    fun findString s t = KMP.find (explode s) (explode t)
    and displayKMP string = KMP.display (explode string)
end;

findString "ab" "adbaabacabba";  (* 0 [4,8] : int list *)
findString "ab" "abaabacabba";   (* = [0,3,7] : int list *)
displayKMP "ab";

displayKMP "abbabbb";


local 
    structure IntKMP = KMPautomaton(IntLabels)
in 
    fun displayKMP (ns : int list) = IntKMP.display ns
    fun findInts s t = IntKMP.find s t
end

findInts [1,2] [1,1,5,0,1,2,1,1];
displayKMP [1,1,5,0,1,1];


Example 2: simplified KMP-automaton: transitions for pattern symbols only

local
    structure sKMP = simpleKMPautomaton(CharLabels)
in
    fun findString s t = sKMP.find (explode s) (explode t)
    and displayKMP s = sKMP.display (explode s)
end

displayKMP "abbabbb";
findString "abbabbb" "abbabbbabbabbb"; (* [0,7] *)
findString "ab" "abaabaabba";          (* [0,3,6] *)
findString "ab" "abacabaabba";         (* [0] wegen c *)


local 
    structure IntsKMP = simpleKMPautomaton(IntLabels)
in
    fun findInts ns ms = IntsKMP.find ns ms 
    and displayIntKMP ns = IntsKMP.display ns
end

findInts [0,0] [0,0,0,1];
displayIntKMP [0,0,1]


REMARK: the KMP-automaton is the minimal counter-DFA accepting L(alphabeth^*;pattern):

structure charMatch = Matcher(CharLabels);
open charMatch;

val pattern = (Times (map Atom (explode "abbabbb")));
display pattern;
find pattern (explode "abbabbbabbabbba")
*)




