(* Matrixalgebra over a polymorphic Kleenean Algebra         H.Leiss, 20.1.99 *)

signature pkaMATRIX = 
    sig
	structure Ring : pKA
	type 'a t                            (* matrices over Ring *)
	val dimension : 'a t -> int * int
	val tabulate : (int * int) -> (int * int -> 'a Ring.t) -> 'a t
	val sub : 'a t * (int * int) -> 'a Ring.t
	val map : ('a Ring.t -> 'a Ring.t) -> 'a t -> 'a t

	val zero : int * int -> 'a t
	val unit : int * int -> 'a t
	val sum  : ''a t * ''a t -> ''a t
	val prod : ''a t * ''a t -> ''a t
	val star : ''a t -> ''a t
	val transpose : ''a t -> ''a t
    end 

functor pkaMatrix (R : pKA) : pkaMATRIX = 
    struct
	structure Ring = R
	type ''a r = ''a Ring.t  

	type 'a t = 'a r vector vector 

	fun dimension (S:'a t) = 
	    let 
		val rows = Vector.length S
		val columns = Vector.length (Vector.sub(S,0)) 
		val samelengths = 
		    Vector.foldr (fn (v,b) => b andalso (Vector.length v = columns)) 
		                 true S
	    in 
		if samelengths then (rows,columns) else raise Size 
	    end

	fun sub (E:'a t,(i,j)) = 
	    Vector.sub (Vector.sub (E,i), j) handle subscript => raise subscript

	fun tabulate (rows,cols) (f : int * int -> 'a r) =
	    Vector.tabulate(rows, fn i => Vector.tabulate (cols, fn k => f(i,k))) :'a t

	fun map (f : 'a r -> 'a r) (A : 'a t) =
	    let 
		val (n,m) = dimension (A:'a t)
	    in
		tabulate (n,m) (fn (i,j) => f (sub (A,(i,j))))
	    end

	fun zero (n,m) = tabulate (n,m) (fn (i,k) => Ring.zero) 
	fun unit (n,m) = tabulate (n,m) (fn (i,k) => if i=k then Ring.unit else Ring.zero) 
	fun transpose (S : ''a t) : ''a t =
	    let 
		val (n,m) = dimension S 
	    in 
		tabulate (m,n) (fn (i,j) => sub(S,(j,i)))
	    end

	fun sum (S:''a t,T:''a t) : ''a t = 
	    let
		val (ns,ms) = dimension S
		val (nt,mt) = dimension T
		fun sumV (v,w) : ''a r vector = 
		    Vector.tabulate (Vector.length v, 
				     fn i => Ring.sum (Vector.sub(v,i), Vector.sub(w,i))) 
	    in
		if (ns,ms)=(nt,mt) then
		    Vector.tabulate (ns, fn i => sumV (Vector.sub(S,i), Vector.sub(T,i)))
		else raise Size
	    end

	fun prod (S:''a t,T:''a t) : ''a t =
	    let 
		val (ns,ms) = dimension S
		val (nt,mt) = dimension T
		fun prodV  (v,w) : ''a r = 
		    let 
			fun sm i = if i=0 then Ring.zero 
				   else Ring.sum(sm (i-1), 
					      Ring.prod (Vector.sub(v,i-1),Vector.sub(w,i-1)))
		    in 
			sm (Vector.length v)
		    end
		fun column (T:''a t,j) = Vector.map (fn r => Vector.sub (r,j)) T
 	    in
		if ms=nt then
		    tabulate (ns,mt) (fn (i,j) => prodV (Vector.sub (S,i), column(T,j)))
		else raise Size
	    end

	fun star (E:''a t) : ''a t =
	    let
		val (d,e) = dimension E
	    in 
		if d=e then
		    if d=1 then tabulate (1,1) (fn (i,j) => Ring.star (sub (E,(i,j))))
		    else
			let 
			    val m = Int.div(d,2) (* recursion on dimension *)

			    val A = tabulate (m,m)     (fn (i,j) => sub (E,(i,j)))
			    val B = tabulate (m,d-m)   (fn (i,j) => sub (E,(i,j+m)))
			    val C = tabulate (d-m,m)   (fn (i,j) => sub (E,(i+m,j)))
			    val D = tabulate (d-m,d-m) (fn (i,j) => sub (E,(i+m,j+m)))

			    (* recursion:  E* = [[F*, F*BD*], [D*CF*, D* + D*CF*BD*]  *)
			    val Ds  = star D
			    val BDs = prod(B,Ds)
			    val F   = sum(A,prod(BDs,C))

			    val A' = star F
			    val B' = prod(A', BDs)
			    val C' = prod(Ds, prod(C,A'))
			    val D' = sum(Ds, prod(C',BDs))
			in
			    tabulate (d,d) (fn (i,j) => 
					    if i < m
						then if j < m then sub (A',(i,j))
						     else sub (B',(i,j-m))
					    else if j < m then sub (C',(i-m,j))
						 else sub (D',(i-m,j-m)))
			end
		else raise Size
	    end
    end

(* Approximatively the matrix algebra over the free Kleenean algebra over 'a *)

structure RegMat : pkaMATRIX = pkaMatrix(Reg) 



