(* File:    ppfm.sml                                                16.02.00
   Content: Functor that extends structure Fm for finite automata by 
            adding display facilities, given print-names for labels.
   Author:  Hans Leiss, CIS, 12/99,  leiss@cis.uni-muenchen.de
   
   ppFm can also display machines constructed with Fm, if the labels fit. 
   This would not work with ppFM(A : LABEL) : (PPFM with type label = A.label)

   Edited:  05.05.08  changed "ghostview --magstep -2 " to "gv --scale=-2 "
*)

functor ppFm(Alphabet : LABEL) : 
  sig
    include FM

    eqtype label (* = Alphabet.label  (* better than PPFM with type label = .. ! *) *)

    val dot : label machine -> string -> unit
    (* write a description of m to the file s in the format of graphviz/dot *)

    val viewgraph : string -> OS.Process.status
    (* display the graph in file s by the graph-drawing function dot. 
       The filename is relative to the directory where SML was called. *)

    val display : label machine -> string -> OS.Process.status
    (* display a picture of the machine m using temporary .dot- and .ps-files with
       basename s. Leaves these files for possible further use. *)

    val displayMinimize : label machine -> string -> OS.Process.status
    (* display a picture of the minimal deterministic machine equivalent to m,
       using temporary .dot- and .ps-files with basename s. 
       Leaves these files for possible further use. *)

    val displayDetermine : label machine -> string -> OS.Process.status
    (* display a picture of the deterministic machine equivalent to m,
       using temporary .dot- and .ps-files with
       basename s. Leaves these files for possible further use. *)

    val displayDetermineMinimize : label machine -> string -> OS.Process.status
        (* a) displays the original automaton m,
           b) displays the deterministic version m2 of m obtained by the powerset construction,
              using sets of states of m to represent states of m2 
           c) displays the deterministic automation m2 with states renamed by numbers
           d) displays the minimal automaton m3 with states as sets of states of m2
           e) displays the minimal automaton equivalent to m with states renamed by numbers.
         *)
    val displayMinimizeHistory : label machine -> string -> OS.Process.status
  end 

= struct
    local 
      val incr = fn (n:int) => (fn (x,a,y) => (x+n,a,y+n))
      val addNat  = fn (n:int) => (fn m => m+n)

      fun member x l = List.exists (fn y => y = x) l
      and add q [] = [q]
        | add q (u::U) = if q=u then (u::U) else (u::(add q U))
      and union [] U = U
        | union (q::Q) U = union Q (add q U)
      and intersects [] Q = false
        | intersects (p::P) Q = if member p Q then true 
                                else intersects P Q
      and Union [] = []
        | Union (Q::Qs) = union Q (Union Qs)
      and difference S [] = S  
        | difference S (e::E) = difference (subtract S e) E
      and subtract [] e = []
        | subtract (s::S) e = if s=e then subtract S e 
                              else (s::(subtract S e))
    in
      open Fm
      type label = Alphabet.label 
 
      fun start M = let val {start = S,...} = show M in S end
      and final M = let val {final = F,...} = show M in F end
      and trans M = let val {trans = T,...} = show M in T end

      fun draw M os name =
        let
          fun write w = TextIO.output (os, w)
          fun drawTransition (i,a,j) =
            write (Int.toString i ^ " -> " ^ Int.toString j 
                   ^ " [label = " ^ (Alphabet.toString a) ^ "];\n")
          fun drawInitialState i = 
            write (Int.toString i ^ " [shape = diamond] ;\n")
          fun drawFinalState i = 
              if (member i (start M)) then 
                  write (Int.toString i ^ " [shape = diamond, peripheries = 2] ;\n")
              else write (Int.toString i ^ " [peripheries = 2] ;\n")
        in 
          write ("digraph " ^ name ^ 
                 " {center=true;\n size=\"7.5,14\";\n ratio=compress;\n");
          app drawTransition (trans M);
          app drawInitialState (start M);
          app drawFinalState (final M);
          write "}\n\n"
        end

      fun drawWithStateSets (M,stateAssoc) os name =
        let
          fun write w = TextIO.output (os, w)
          fun node i = "node"^Int.toString i
          fun drawTransition (i,a,j) = write (node i ^ " -> " ^ (node j) ^ 
                                       " [label = " ^ (Alphabet.toString a) ^ "];\n")
          fun drawInitialState i = write (node i ^ " [shape = diamond];\n")
          fun drawFinalState i =   
              if (List.exists (fn x => x=i) (start M)) 
                  then write (node i ^ " [shape = diamond, peripheries = 2];\n")
              else write (node i ^ " [peripheries = 2];\n")

          fun defineNode (i,oldstates) =   
              write (node i^"[label = \""^(toString oldstates)^"\"];\n")
          and toString (p::q::rest) = Int.toString p^","^(toString (q::rest))
            | toString (p::nil) = Int.toString p
            | toString nil = " "
        in 
          write ("digraph " ^ name ^ 
                 " {center=true;\n size=\"7.5,14\";\n ratio=compress;\n");
          app defineNode stateAssoc;
          app drawTransition (trans M);
          app drawInitialState (start M);
          app drawFinalState (final M);
          write "}\n"
        end

      fun dot M filename =
        let
          val os = TextIO.openOut (filename ^ ".tmp.dot")
        in
          draw M os filename;
          TextIO.closeOut os
        end
      and dotDetermineMinimize M filename =
        let
          val os = TextIO.openOut (filename ^ ".tmp.dot")
          val (Mdet,stateAssocD) = determineWithStateAssoc M
          val (Mmin,stateAssocM) = minimizeWithStateAssoc Mdet
        in
          draw M os ("PartialNonDet");
          drawWithStateSets (Mdet,stateAssocD) os "DetWithStateSets";
          draw Mdet os "Det";
          drawWithStateSets (Mmin,stateAssocM) os "MinWithStateSets";
          draw Mmin os "MinDet";
          TextIO.closeOut os
        end
      and dotDetermine M filename =
        let
          val os = TextIO.openOut (filename ^ ".tmp.dot")
          val (Mdet,stateAssoc) = determineWithStateAssoc M
        in
          draw M os (filename^"_PartialNonDet");
          drawWithStateSets (Mdet,stateAssoc) os "DetWithStateSets";
          draw Mdet os "Det";
          TextIO.closeOut os
        end
      and dotMinimize M filename = (* M must be total deterministic *)
        let
          val os = TextIO.openOut (filename ^ ".tmp.dot")
          val (Mmin,stateAssoc) = minimizeWithStateAssoc M
        in
          draw M os ("TotalDet");
          drawWithStateSets (Mmin,stateAssoc) os "MinDetWithStateSets";
          draw Mmin os "MinDet";
          TextIO.closeOut os
        end

      fun viewgraph filename =
        (OS.Process.system 
         ("dot -Tps " ^ filename ^ ".tmp.dot -o " ^ filename ^ ".tmp.ps ");
         (* OS.Process.system ("kghostview " ^ filename ^ ".tmp.ps &")) *)
         (* OS.Process.system ("ghostview -magstep -2 " ^ filename ^ ".tmp.ps &") *)
         OS.Process.system ("gv --scale=-2 " ^ filename ^ ".tmp.ps &"))

      fun displayMinimize M filename = 
        (dotMinimize M filename; viewgraph filename)
      and displayDetermine M filename =
        (dotDetermine M filename; viewgraph filename)
      and displayDetermineMinimize M filename =
        (dotDetermineMinimize M filename; viewgraph filename)

      fun display M filename = (dot M filename; viewgraph filename)
      
      fun displayMinimizeHistory M filename =
        (dotMinWithHistory M filename; viewgraph filename)
      and dotMinWithHistory M filename =
        let
          val os = TextIO.openOut (filename ^ ".tmp.dot")
          val ApproxMs = minimizeWithHistory M 

          fun drawMinWithHistory os ((M,Assoc) :: rest) n =
              (drawWithHistory (M,Assoc) os ("ApproxMin" ^ (Int.toString n));
               drawMinWithHistory os rest (n-1))
            | drawMinWithHistory os [] n = TextIO.output (os, "\n\n")
        in
          drawMinWithHistory os ApproxMs (length ApproxMs);
          TextIO.closeOut os
        end
      and drawWithHistory (M,Assoc) os name =
        let
          val {start = Start, trans = Trans, final = Final} = show M
          fun write w = TextIO.output (os, w)

          fun node i = "node"^Int.toString i
          and defineNode (i,oldstates) =   
              write (node i^"[label = \""^(toString oldstates)^"\"];\n")
          and toString (p::q::rest) = Int.toString p^","^(toString (q::rest))
            | toString (p::nil) = Int.toString p
            | toString nil = " "

          and drawTransition (i,(a,ints),j) =
            write (node i ^ " -> " ^ node j 
                   ^ " [label = \" " ^ (Alphabet.toString a) ^ "; " ^ (toString ints) ^"\"];\n")

          fun drawInitialState i = write (node i ^ " [shape = diamond] ;\n")
          fun drawFinalState i = 
              if (member i Start) then 
                  write (node i ^ " [shape = diamond, peripheries = 2] ;\n")
              else write (node i ^ " [peripheries = 2] ;\n")
        in 
          write ("digraph " ^ name ^ 
                 " {center=true;\n size=\"7.5,14\";\n ratio=compress;\n");
          app defineNode Assoc;
          app drawTransition Trans; 
          app drawInitialState Start;
          app drawFinalState Final;
          write "}\n\n"
        end

    end
  end;

(* 

structure iM = ppFm(IntLabels);
structure sM = ppFm(StringLabels); (* : PPFM transparent type label = string! *)
*)

(* Compiler.Control.Print.printDepth := 10;             

sM.displayDetermine (Fm.Star M) "test"
sM.displayDetermineMinimize (Fm.Star M) "test"

Example 0: (Unreachable states are not eliminated)

val M = iM.make{final=[0],start=[0],trans=[(0,1,0),(0,0,0),(1,0,1),(1,1,0)]}
iM.show(iM.minimize M);
val it = {final=[0],start=[0],trans=[(0,1,0),(0,0,0),(1,0,1),(1,1,0)]}
iM.displayMinimize M "test"

Example 1 (Minimize a total deterministic automaton) (Skriptum S.46) 

val M = sM.make{
trans = [(1,"a",2),(1,"b",5),(2,"a",4),(2,"b",3),(3,"a",2),(3,"b",4),
         (4,"a",4),(4,"b",5),(5,"a",4),(5,"b",5)],
start = [1],
final = [4,5]}

sM.show(sM.minimize M);
val it =
  {final=[0],
   trans=[(1,"b",0),(2,"b",1),(0,"a",0),(0,"b",0),(2,"a",0),(1,"a",2)],
   start=[1]}
sM.minimizeWithStateAssoc M;
val it = (-,[(1,[1,3]),(2,[2]),(0,[5,4])])

sM.displayMinimize M "M49";
sM.displayMinimizeHistory M "M49d";


Example 2 (Minimize a total deterministic automaton with unreachable states)
(Skriptum S.50: a=1,...,h=8, 0=a,1=b)

val M = sM.make{
start = [1],
final = [3], 
trans = [(1,"a",2),(1,"b",6),(2,"a",7),(2,"b",3),
         (3,"a",1),(3,"b",3),(4,"a",3),(4,"b",7),
         (5,"a",8),(5,"b",6),(6,"a",3),(6,"b",7),
         (7,"a",7),(7,"b",5),(8,"a",7),(8,"b",3)]};
sM.display M "M50";

val (Md,EquivStates) = sM.determineWithStateAssoc M;
val EquivStates = [(0,[1]),(1,[6]),(2,[2]),(3,[7]),(4,[3]),(5,[5]),(6,[8])]
sM.show Md;
val it =
  {final=[4],
   trans=[(0,"b",1),(0,"a",2),(1,"b",3),(1,"a",4),(2,"b",4),(2,"a",3),
          (3,"b",5),(3,"a",3),(4,"b",4),(4,"a",0),(5,"b",1),(5,"a",6),
          (6,"b",4),(6,"a",3)],
   start=[0]}

sM.minimizeWithStateAssoc Md;
val it = (-,[(4,[3]),(2,[2,6]),(3,[1]),(1,[0,5]),(0,[4])])

sM.displayMinimize Md "M51d";
a) show determinstic automaton,
b) show minimal automaton with state sets of deterministic automaton
c) show minimal automaton with renamed states.

sM.displayDetermineMinimize M "M51dm"
a) show original automaton,
b) show deterministic version with state sets (here: remove unreachable state 4)
c) show deterministic automaton with states renamed
d) show minimal automaton with state sets of renamed deterministic automaton
e) show minimal automaton with renamed states.

sM.displayMinimizeHistory Md "M51dmh"  (* ins Skriptum ?? *)


Example 3 (Minimize a partial non-deterministic automaton) 
          (Rem. minimizaton fails, unless we first make M deterministic)

val M = iM.make{
start = [1,3],
final = [3], 
trans = [(1,0,2),(1,0,6),(2,0,7),(2,1,3),(3,0,1),(3,1,3),(4,0,3),(4,1,7)]};
iM.minimize M;
iM.displayMinimize M "Mkn"

iM.displayDetermineMinimize M "M50";
val (Mdet,Equiv) = iM.determineWithStateAssoc M

val M = 
iM.make{final=[0,1],start=[0],
        trans=[(0,1,1),(0,0,2),(1,1,1),(1,0,3),(2,1,1),(2,0,4),(3,1,5),(3,0,6),
               (4,1,1),(4,0,7),(5,1,5),(5,0,5),(6,0,7),(6,1,1),(7,0,5),(7,1,5)]}
iM.accept M [1,0,1];
val it = false : bool

val (Mmin,Equiv) = iM.minimizeWithStateAssoc M
val Equiv = [(0,[0]),(5,[2]),(3,[1]),(4,[3]),(1,[4,6]),(2,[5,7])]
iM.show Mmin;
val it =
  {final=[0,3],start=[0],
   trans=[(5,0,1),(4,0,1),(2,1,2),(1,0,2),(2,0,2),(1,1,3),(3,1,3),(0,1,3),
          (4,1,2),(3,0,4),(5,1,3),(0,0,5)]}

iM.displayMinimize M "M";
iM.displayMinimizeHistory M "M";  (* true minimal automaton with  4={4,6}, 5={5,7} *)


val Mb = 
iM.make{final=[1],start=[0],
        trans=[(0,1,1),(0,0,2),(1,1,1),(1,0,3),(2,1,1),(2,0,4),
               (3,1,3),(3,0,4),(4,1,1),(4,0,4)]};

val (Mbmin,Equiv) = iM.minimizeWithStateAssoc Mb;
iM.show Mbmin
val it =
  {final=[0],start=[1],trans=[(1,1,0),(1,0,1),(0,1,0),(2,0,1),(2,1,2),(0,0,2)]}


------------------------------------------------------------------------------

Example 4: (Deterministic version can have exponential size)

Make automata [corresponding to (a+b)^*a(a+b)^n] deterministic and
show the states as sets of old states.  (For the translation from
regular expressions (a+b)^*a(a+b)^n to finite automata, see
regaut.sml.)

val M0 = iM.make{
start = [0],
final = [1], 
trans = [(0,0,0),(0,1,0),(0,0,1)]};

iM.displayDetermine M0 "M0Det";

val M1 = iM.make{
start = [0],
final = [2], 
trans = [(0,0,0),(0,1,0),(0,0,1),(1,0,2),(1,1,2)]};

iM.displayDetermine M1 "M1det";
iM.displayDetermineMinimize M1 "M1detmin";

val M2 = iM.make{
start = [0],
final = [3], 
trans = [(0,0,0),(0,1,0),(0,0,1),(1,0,2),(1,1,2),(2,0,3),(2,1,3)]};

iM.displayDetermine M2 "M2det";
iM.displayMinimize M2 "M2min";
iM.displayDetermineMinimize M2 "M2detmin";

val M3 = iM.make{
start = [0],
final = [4], 
trans = [(0,0,0),(0,1,0),(0,0,1),(1,0,2),(1,1,2),(2,0,3),(2,1,3),
         (3,0,4),(3,1,4)]};
iM.displayDetermine M3 "M3det";

val M4 = iM.make{
start = [0],
final = [5], 
trans = [(0,0,0),(0,1,0),(0,0,1),(1,0,2),(1,1,2),(2,0,3),(2,1,3),
         (3,0,4),(3,1,4),(4,0,5),(4,1,5)]};
iM.displayDetermine M4 "M4det";



val M = sM.make{
start = [0],
final = [2], 
trans = [(0,"a",1),(0,"b",2),(1,"a",2),(2,"b",0),(1,"b",0)]}

sM.displayMinimizeHistory M "test";

*)
