(* File:    pattern-matcher.sml
   Content: - functor Matcher : LABEL -> PPMATCHER that generates from an alphabet
              a structure with three functions:
              - matcher: turns a regular expression (defining a 
                regular set of pattern sequences) into a (minimal) finite automaton 
                with a counter to remember the (end position of) occurrences of 
                patterns in an input sequence.
              - display: displays the pattern-match automaton for expression r
              - find: searches for pattern occurrences and reports occurrences 
                in an input by their end positions.
   Author:  Hans Leiss, CIS, leiss@cis.uni-muenchne.de
   Date:    16.12.99,  28.12.99
   Edited:
   24.02.00 - separated from counterfm.sml
*)

signature PPMATCHER =
  sig
    eqtype t 
    datatype s = Dummy | Pat of t

    datatype re = datatype RegExp.re

    structure M : PPCFM where type label = s
    val matcher : t re -> s M.machine
    val display : t re -> OS.Process.status
    val find : (t re) -> (t list) -> int list
  end

functor ppMatcher(Alphabet : LABEL) :> PPMATCHER where type t = Alphabet.label = 
    struct 
        type t = Alphabet.label
        (* open Alphabet :> LABEL  to provide datatype constructors of Alphabet.label ?? *)

        datatype s = Pat of t | Dummy       (* for edges with labels not in pattern *)
        datatype re = datatype RegExp.re

        structure NewAlphabet = 
            struct 
                type label = s 
                fun toString Dummy = "_"  (* for .dot-drawings *)
                  | toString (Pat(t)) = Alphabet.toString t
            end

        structure M = ppCFm(NewAlphabet) : PPCFM (* finite automata with counter *)
        structure RegFm = ppRegExpFm(NewAlphabet) 

        fun pats (Atom a) acc = add a acc
          | pats (Var i) acc = acc
          | pats (Plus []) acc = acc
          | pats (Plus (r::rs)) acc = pats (Plus rs) (pats r acc)
          | pats (Times []) acc = acc
          | pats (Times (r::rs)) acc = pats (Times rs) (pats r acc)
          | pats (Star r) acc = pats r acc
        and add x [] = [x]
          | add x (y::xs) = if x=y then (y::xs) else (y::(add x xs))

        fun matcher(r : t re) =
            let 
                val patternsymbols = pats r []
                val labels = (Dummy :: (map (fn t => Pat(t)) patternsymbols))

                fun substatoms (Atom a) = Atom (Pat a)
                  | substatoms (Var i) = Var i
                  | substatoms (Plus rs) = Plus (map substatoms rs)
                  | substatoms (Times rs) = Times (map substatoms rs)
                  | substatoms (Star r) = Star (substatoms r)

                val pM = RegFm.retominfm (Times[Star(Plus (map Atom labels)),substatoms r])
            in
                M.make(RegFm.M.show pM)
            end

        fun display (r : t re) = M.display (matcher r) "matcher"

        fun find (pattern : t re) (text : t list) =
            let 
                val PM = matcher pattern
                val patternsymbols = pats pattern []
                fun convert x = if List.exists (fn y => y=x) patternsymbols 
                                   then Pat(x) else Dummy
            in
                List.rev (M.hits PM (map convert text))
            end
    end;


(* Example 1:

structure Match = ppMatcher(StringLabels) ;
open Match;

display(Atom "a");
display(Times[Atom "a",Atom "b"]);
display(Times[]); (* falsch *)

display(Plus[Times[Atom "a",Atom "c"],Atom "d"]);

val patM = (Times [Atom "a",Star(Atom "a")]);
display patM;
find patM ["a","b","a","a","cd","a"]

Example 2: 

structure charMatch = ppMatcher(CharLabels);
open charMatch;

find (Times [Atom #"a",Star(Atom #"a")]) (explode "abaaba");
display (Times [Atom #"a",Star(Atom #"a")]);

fun stringToCharRe s = Times (map Atom (explode s));

display (Plus (map stringToCharRe ["ab","ba"]));
find (Plus (map stringToCharRe ["ab","ba"])) (explode "abaab");

display (stringToCharRe "abbabbb");
find (Times[]) (explode "abcab");  (* Position 0 not given: [1,2,3,4,5] *)

*)




