(* File:    ppregexp.sml
   Content: functor creating printable regular expressions and equation systems
   Date:    12/99
   Author:  H.Leiss
   Status:  preliminary 
   Edited:
   15.02.00 renamed ppREGEXP to ppRegExp
   02.10.09 renamed Compiler.PrettyPrint to PrettyPrint
            (since config/preloads autoloads $smlnj/compiler.cm, the visible 
             compiler is a collection of structures, not Compiler),
   04.01.10 adapted pp-functions to PrettyPrintNew, removed en_pp 
*)

signature PPREGEXP =
  sig
    include REGEXP_WEAK

    type 'a eqn = int * 'a re  
    (* recursion equation between variable and expression *)
    type 'a eqns = 'a eqn list 
    (* system of regular recursion equations, EBNF-grammar *)

    eqtype alph 
    (* alphabet of constants in regular expressions *)

    val printEqns : alph eqns -> unit
    val printExp : alph re -> unit
  end

functor ppRegExp (N : NAMES) : PPREGEXP where type alph = N.alph =
    let
        structure PP = PrettyPrint

        val defaultDevice : PP.device = {consumer = Control.Print.say,
                                         linewidth = !Control.Print.linewidth,
                                         flush = Control.Print.flush}
        val ppDefault = PP.with_pp defaultDevice
    in
        struct
	    open RegExp (* : REGEXP_WEAK by PPREGEXP *)
	    (* datatype re = datatype RegExp.re *)

            type alph = N.alph
	    type exp = alph re	

            val varName = N.varName    (* names of variables, provided by the user *)
	    and atomName = N.atomName  (* names of atoms,     provided by the user *)

            fun ppExp (env : StaticEnv.staticEnv) ppstrm (e : exp) = 
                let 
                    val pps = PP.string ppstrm
                in 
                    case e of
                        (Atom a) => pps (atomName a)
                      | (Var i)  => pps (varName i)
                      | (Plus []) => pps "0"
                      | (Plus [e]) => ppExp env ppstrm e
                      | (Plus (e::es)) => (pps "("; ppExp env ppstrm e;
                                           ppArgs env ppstrm (" + ",es); pps ")")
                      | (Times []) => pps "1"
                      | (Times (e::es)) => (ppExp env ppstrm e; 
                                            ppArgs env ppstrm (";",es))
                      | (Star (Star e)) => (ppExp env ppstrm (Star e); pps "*")
                      | (Star (Plus es)) => (ppExp env ppstrm (Plus es); pps "*")
                      | (Star e) => (pps "("; ppExp env ppstrm e; pps ")*")
                end
            and ppArgs env ppstrm (infx,es) : unit = 
                (PP.openHOVBox ppstrm (PP.Rel 1) ; 
                 ppArgs1 env ppstrm (infx,es); 
                 PP.closeBox ppstrm               
                )
            and ppArgs1 env ppstrm (infx,es) = 
                let 
                    val pps = PP.string ppstrm
                    fun ppArg e = (pps infx; ppExp env ppstrm e)
                in 
                    case es of [e] => ppArg e
                  | (e::rs) => (ppArg e; ppArgs1 env ppstrm (infx,rs))
                  | _ => ()
                end

            fun prExp env exp = ppDefault (fn ppstream => ppExp env ppstream exp)
            and printExp exp = prExp StaticEnv.empty exp

            type 'a eqn = (int * 'a RegExp.re)
            type 'a eqns = 'a eqn list

            fun ppEqns (env : StaticEnv.staticEnv) ppstrm (eqs : alph eqns) : unit = 
                let 
                    val pps = PP.string ppstrm
                in 
                    (PP.openHOVBox ppstrm (PP.Rel 1); 
                     PP.newline ppstrm;
                     pps "Equation system: ";
		     pps "\n"; 
                     ppEqns1 env ppstrm eqs;
                     PP.closeBox ppstrm)
                end
            and ppEqns1 env ppstrm eqs = 
                let 
                    val pps = PP.string ppstrm
                in 
                    app (fn ((i,e): alph eqn) =>
                         (PP.openHVBox ppstrm (PP.Rel 0); 
                          pps "  ";
                          ppExp env ppstrm (Var i); 
                          pps " = ";
                          ppExp env ppstrm e;
                          pps "\n";
                          PP.closeBox ppstrm               
                        )) eqs
                end

            fun prEqns env eqs = ppDefault (fn ppstream => ppEqns env ppstream eqs)
            and printEqns eqs = prEqns StaticEnv.empty eqs
        end
    end


(* generic Example: -------------------------------------------------------- 

structure R = ppRegExp(Names : NAMES);
R.printExp r;
R.printEqns reqns;

--------- *)


(* Example: -------------------------------------------------------- 

structure R = 
    let 
	structure Names =
	    struct
		type alph = string
		fun atomName (a : alph) = a : string
		fun varName i = if i=6 then "x" 
				else "y("^ Int.toString i ^")"
	    end
    in 
	ppRegExp (Names)
    end;

let
    open R
in
    printExp (Star (Plus[Atom "a",
			 Star (Star(Times[Var 8,Atom "B"])), 
			 Star(Var 6),
			 Atom "e"]));
    printEqns [(1,Star (Plus [Atom "ab",Var 0])),
	       (0,Times[Atom "a",Atom "ba",Plus [Var 0,Var 1]]),
	       (2,Times[Atom "aba",Plus [Var 10]])]
end;


------------------------------------------------------------------ *)
