(* File: Automata-Lib/regeqns.sml

   12.02.00 - moved Reg : pKA to Automata-Lib/pka.sml, 
              moved RegMat : pkaMATRIX to regmat.sml        

From regSys.sml:

 TO BE DONE:
   - allow free variables in equation systems; this makes everything
     more complicated, but more rigorous and flexible as well.

   - check rightFactorize: expressions seem too complicated;
     was better with binary Plus,Times. Done??

   - check CFG for e-freeness etc in GNF-computation.
*)

signature REGEQNS =
    sig	
	include REGEXP_WEAK
        
        type 'b eqns = (int * 'b) list
        type 'a env  = (int -> 'a)

	val maxVar   : 'a re eqns -> int
	val nthVar   : 'a re eqns * int -> 'a re
	val freeVars : 'a re eqns -> int list
	val recVars  : 'a re eqns -> int list

	val eliminateStars : ''a re eqns -> ''a re eqns
	val convertToCFG   : ''a re eqns -> ''a re eqns
	val greibachNF     : ''a re eqns -> ''a re eqns
	val leftToRightRecursion  : ''a re eqns -> ''a re eqns
	val isolateLeftVarFactors : ''a re eqns -> ''a RegMat.t * ''a RegMat.t 
	val solveRightlinearEqns  : ''a re eqns -> ''a re eqns

	val fix   : ('a env -> 'b -> 'a) -> 'a env -> ('a env -> ''c) 
                                                   -> 'b eqns -> 'a env
	val solve : ('a env -> 'b -> 'a) -> 'a env -> ('a env -> ''c) 
                                                   -> 'b eqns -> 'a eqns
	val iter    : ('a env -> 'b -> 'a) -> 'a env -> int -> 'b eqns -> 'a env 
	val iterate : ('a env -> 'b -> 'a) -> 'a env -> int -> 'b eqns -> 'a eqns
     end

structure RegEqns : REGEQNS =
    struct
	open RegExp 
        type 'b eqns = (int * 'b) list

	exception Undefined

	fun maxVar eqns = Int.max(maxRecVar eqns, maxParVar eqns)
	and maxRecVar eqns = foldl (fn ((i,r),j) => Int.max(i,j)) ~1 eqns
	and maxParVar eqns = 
	    foldl (fn ((i,r),j) => Int.max(RegExp.maxVar r,j)) ~1 eqns

        fun recVars (eqs : 'a re eqns) = map #1 eqs
        and freeVars (eqs : 'a re eqns) = 
            Set.list (Set.difference
                      (Set.Union (map (fn (i,e) => RegExp.freeVars e) eqs))
                      (Set.mk (recVars eqs)))

	fun nthVar (eqns,i) = 
	    case List.nth (eqns,i) of (j,r) => Var j 
	fun varnum (Var k) = k | varnum _ = raise Undefined

	fun eliminateStars eqns = 
	    let 
		fun findStar (Var i) = Set.empty
		  | findStar (Atom a) = Set.empty
		  | findStar (Plus rs) = Set.Union (map findStar rs)
		  | findStar (Times rs) = Set.Union (map findStar rs)
		  | findStar (Star r) = Set.mk [r]

		val starred = Set.list (Set.Union (map (findStar o #2) eqns))

		val enumStarred = 
		    let  
			fun enum [] i = []
			  | enum (s::ss) i = (i,s)::(enum ss (i+1))
			val m = 1 + maxVar eqns
		    in 
			enum starred m
		    end

		val starEqns = 
		    map (fn (i,s) => (i, Plus[Times[s,Var i],Unit])) enumStarred

		fun substStarred (Var i) = (Var i)
		  | substStarred (Atom a) = (Atom a)
		  | substStarred (Plus rs) = Plus (map substStarred rs)
		  | substStarred (Times rs) = Times (map substStarred rs)
		  | substStarred (Star r) = 
		    case List.find (fn (i,s) => s=r) enumStarred of
			SOME (i,s) => (Var i) | NONE => (Star r)

		val topStarsReplaced = 
		    (map (fn (i,r) => (i,substStarred r)) eqns) @ starEqns
	    
	    in 
		if starred = [] then eqns else
		    eliminateStars (topStarsReplaced)
	    end

	fun convertToCFG eqns =
	    map (fn (i,r) => (i,dnf r)) (eliminateStars eqns)

	fun solveRightlinearEqns eqns =
	    let 
		val (A,B) = isolateRightVarFactors eqns
		val X = RegMat.prod(RegMat.star A,B)
	    in 
		List.tabulate (length eqns, 
			       fn k => (varnum (nthVar(eqns,k)), RegMat.sub(X,(k,0))))
	    end
	and isolateRightVarFactors eqns = 
	    let 
		val rs = map (fn (i,r) => factorizer r) eqns

		fun A (i,j) = 
		    let 
			fun recVarEnding j ((Times[s,t])::ss) =
			    if t = nthVar (eqns,j) then s else (recVarEnding j ss)
			  | recVarEnding j (_ :: ss) = raise Undefined
			  | recVarEnding j [] = Plus []
		    in
			case List.nth (rs,i) of
			    Plus ss => recVarEnding j ss
			  | _ => raise Undefined
		    end

		fun B (i,j) = 
		    let 
			fun isRecVar (Var k) = 
			    List.exists (fn j => j=k) (recVars eqns)
			  | isRecVar _ = false
			fun constLeading ((Times[s,t])::ss) acc = 
			    if isRecVar t then constLeading ss acc
			    else (Times[s,t])::acc
			  | constLeading (_ :: ss) acc = raise Undefined
			  | constLeading [] acc = acc
		    in
			case List.nth (rs,i) of
			    Plus ss => Plus (rev (constLeading ss []))
			  | _ => raise Undefined
		    end

		val n = length eqns 
	    in 
		(RegMat.tabulate (n,n) A, RegMat.tabulate (n,1) B)
	    end

	fun isolateLeftVarFactors eqns = 
	    let 
		val n = length eqns 
		fun isRecVar (Var k) = recVar k eqns
		  | isRecVar _ = false
		and recVar k ((i,r)::eqs) = (i = k) orelse recVar k eqs
		  | recVar k [] = false

		val rs = map (fn (i,r) => factorizel r) eqns

		fun A (i,j) = 
		    let 
			fun findFactor i ((Times (e::es))::exprs) = 
			    if e = nthVar (eqns,i) then nf(Times es) 
			    else findFactor i exprs
			  | findFactor i (_ :: exprs) = raise Undefined
			  | findFactor i [] = Zero
		    in 
			case (List.nth (rs,j)) of 
			    (Plus ss) => findFactor i ss
			  | _ => raise Undefined
		    end

		fun B (i,j) = 
		    let 
			fun constLeading ((Times (s::ss))::es) acc = 
			    if isRecVar s then (constLeading es acc)
			    else (constLeading es (Times(s::ss)::acc))
			  | constLeading (_ :: es) acc = raise Undefined
			  | constLeading [] acc = acc
		    in
			case List.nth (rs,j) of
			    Plus ss => nf(Plus (constLeading ss []))
			  | _ => raise Undefined
		    end
				    
	    in 
		(RegMat.tabulate (n,n) A, RegMat.tabulate (1,n) B)
	    end

        (* Extract the left recursion in an equation system x = r(x) by
           writing it as a matrix equation x = x;R + t, whose least solution
           is the same as that for the system x = t;R* = t;R+ + t, which is 
           right-recursive when written with additional unknowns Y as

	       x = t;Y + t, Y = R;Y + R.      (R,t depend on x)

           Assuming we have removed epsilon-rules, t has no 1 and hence all 
           remaining leading recursion variables on rhs are from x in R: 
           the new system has no left recursion. 
	*)

	fun leftToRightRecursion eqns =
	    let 
		val (R,T) = isolateLeftVarFactors eqns
		val n = length eqns
		val m = 1 + (maxVar eqns)
		val Y = RegMat.tabulate (n,n) (fn (i,j) => Var (m + n*i +j))

		fun neweqns k = 
		    let
			val X = RegMat.map dnf (RegMat.sum(RegMat.prod(T,Y),T))
			val Z = RegMat.map dnf (RegMat.sum(RegMat.prod(R,Y),R))
		    in
			if k<n then (varnum (nthVar (eqns,k)), RegMat.sub(X,(0,k)))
			else (k, RegMat.sub(Z, (Int.div(k-n,n),Int.rem(k-n,n))))
		    end
	    in
		List.tabulate (n + n*n, neweqns)
	    end


         (* Eliminate leftrecursion in x = r(x) by going to 

	       x = t;Y + t, Y = R;Y + R.

           Avoid variables on the left by substituting (t;Y + t) for x in R.
           Assumes that 1 is eliminated from eqns, so t has no 1's.         *)

	fun greibachNF eqns = 
	    let 
		val (R,T) = isolateLeftVarFactors eqns
		val n = length eqns
		val m = 1 + (maxVar eqns) 
		val Y = RegMat.tabulate (n,n) (fn (i,j) => Var (m + n*i + j))

		val X = RegMat.sum(RegMat.prod(T,Y),T)
		val Z = RegMat.sum(RegMat.prod(R,Y),R)

		val newEs = map (fn (i,r) => (i, RegMat.sub (X,(0,i)))) eqns
		val XZ = RegMat.map (dnf o (subst newEs)) Z

                fun neweqns k = if k < n then (varnum (nthVar (eqns,k)), RegMat.sub(X,(0,k)))
                                else (k, RegMat.sub(XZ, (Int.div(k-n,n), Int.rem(k-n,n))))
	    in
		List.tabulate (n + n*n, neweqns)
	    end

(*        fun ppRegMat prVar A = 
            Vector.map (fn v => Vector.map (RegExp.pp prVar o RegExp.nf) v) A
        and pp (prVar : int -> string) es = 
            List.map (fn (i,r) => ((RegExp.pp prVar (RegExp.Var i))
                                   ^" = "^(RegExp.pp prVar r))) es
*)

	(* Interpretation of recursion equation systems by least fixed points *)

        type 'a env = (int -> 'a)

        fun solve (eval : 'a env -> 'b -> 'a) (init : 'a env) 
            (test : 'a env -> ''c) (eqs: 'b eqns) =
            let 
                val Env = fix eval init test eqs
            in 
		map (fn (i,e) => (i,Env i)) eqs 
            end

        and fix (eval : 'a env -> 'b -> 'a) (init : 'a env) 
            (test : 'a env -> ''c) (eqs: 'b eqns) =
            let 
                fun F Env = let 
                                fun Env' n = 
                                    case (List.find (fn (i,r) => i=n) eqs) of
                                        SOME (i,r) => (eval Env r)
                                      | NONE => Env n
                            in 
                                if (test Env' = test Env) then Env' else (F Env')
                            end
            in 
                F init
            end : 'a env

        (* ----- n-fold iteration of a function f applied to initial values ------ *)

        fun iter (eval : 'a env -> 'b -> 'a) (init : 'a env) n (eqs : 'b eqns) =
	    let
		fun F k Env = let 
				  fun Env' j = 
				      case (List.find (fn (i,r) => i=j) eqs) of
					  SOME (i,r) => (eval Env r)
					| NONE => Env j
			      in 
				  if k <= 0 then Env else 
				      if k = 1 then Env' else (F (k-1) Env')
			      end
	    in
		F n init 
	    end
	and iterate f init n eqs = 
	    let 
		val Env = iter f init n eqs 
	    in 
		map (fn (i,e) => (i,Env i)) eqs
	    end

    end


(* Examples:

open RegEqns;
Compiler.Control.Print.printDepth := 10;

1. Replace left-recursion by right-recursion:

val E = [(0,Plus[Times[Var 1,Atom "a",Var 0],Atom "b"]),
         (1,Plus[Atom "c",Var 0])];
val eLR = leftToRightRecursion E;
val eLR =
  [(0,Plus [Atom "b",Times [Atom "b",Var 2],Times [Atom "c",Var 4]]),
   (1,Plus [Atom "c",Times [Atom "b",Var 3],Times [Atom "c",Var 5]]),(2,Var 4),
   (3,Plus [Var 5,Times []]),
   (4,Plus [Times [Atom "a",Var 0,Var 2],Times [Atom "a",Var 0]]),
   (5,Times [Atom "a",Var 0,Var 3])] : string re eqns

2. Solve a right linear system by reg.exprs: 
val rlE = [(0,Plus[Times[Atom "a",Var 0],Atom "b"]),
           (1,Plus[Times[Atom "c",Var 0],Var 0,Var 1])];

solveRightlinearEqns rlE;
val it =
  [(0,Times [Star (Atom "a"),Atom "b"]),
   (1,Times [Plus [Times [],Atom "c"],Star (Atom "a"),Atom "b"])]
  : string re eqns

3. ConvertToCFG:


val EBNF = [(0,Plus[Star(Times[Var 1,Atom "a",Var 0]),Atom "b"]),
            (1,Times[Plus[Atom "c",Var 0],Var 3])];

convertToCFG EBNF;
val it =
  [(0,Plus [Var 4,Atom "b"]),
   (1,Plus [Times [Atom "c",Var 3],Times [Var 0,Var 3]]),
   (4,Plus [Times [Var 1,Atom "a",Var 0,Var 4],Times []])] : string re eqns


4. 

*)
