(* File:    regeqnsfm.sml 
   Content: Translations from regular expressions to finite automata and conversely
   Date:    12/99
   Author:  H.Leiss

   Edited:
   4.1.99
   - added Example on translation
   12.2.00 moved RegEqnsFm to a new file, for .mldoc
   01.03.00 Renamed file from regexpfm.sml to regeqnsfm.sml,
            extracted the display-functions to ppregeqnsfm.sml
            added fmToEqns to REGEQNSFM
*)

signature REGEQNSFM =
    sig
	structure E : REGEQNS
	structure M : FM 

	type ''a machine = ''a M.machine
        datatype re = datatype E.re

	val accept : ''a machine -> ''a list -> bool
	val retofm : ''a re -> ''a machine
	val retominfm : ''a re -> ''a machine
        val fmtore : ''a machine -> ''a re
        val fmToEqns : ''a machine -> ''a re E.eqns
    end 

(* Allow E to be an extension of RegEqns, intended for ppRegEqns *)

functor RegEqnsFm(structure E : REGEQNS structure M : FM) : REGEQNSFM =
    struct
        structure E = E
        datatype re = datatype E.re

	structure M = M
	type ''a machine = ''a M.machine

	val accept = M.accept

        exception Undefined 

        (* Translation of Regular Expressions to Finite Automata *)

	fun retofm (Atom a) = M.Atom a
          | retofm (Var n) = raise Undefined
	  | retofm (Plus rs) = M.Plus (map retofm rs)
	  | retofm (Times rs) = M.Times (map retofm rs)
	  | retofm (Star r) = M.Star (retofm r)
        and retominfm r = M.minimize (M.determine (retofm r))

        (* Translation of Finite Automata to Regular Expressions *)

	fun fmtore M = 
            let 
                val eqs = E.solveRightlinearEqns (fmToEqns M)
                val eqn = List.find (fn (i,r) => i= M.states M) eqs
            in
                case eqn of SOME(i,r) => r | NONE => raise Undefined
            end

        and fmToEqns M = 
            (* Represent a finite automaton A as a system of right-linear equations

                  { xi = r_i(x1,...,xm) | i=1,...,m } 

              where the variables xi (i<m) are the states of A and ri is a sum with 
              summands of the form (a+b+...)xj or 1. An additional main variable 
      
                    xm = xi + ...+ xk    
  
              is added, where xi,...,xk is the set of initial states of A. Solve this
              to get a regular expression for the language accepted by A 
            *)
            let
                val newStart = (M.states M) 
                val {start = Start, final = Final, trans = Trans} = M.show M

                fun summands k ((i,a,j) :: I) = 
                    if k=i 
                        then (Times[Atom a, Var j] :: (summands k I)) 
                    else (summands k I)
                  | summands k [] = []

                fun rexp i = if member i Final
                                 then Plus (E.Unit :: (summands i Trans))
                             else Plus (summands i Trans)

                fun eqns 0 = [(0, rexp 0)]
                  | eqns n = if n = newStart 
                                 then (n, Plus (map (fn j => Var j) Start)) :: (eqns (n-1))
                             else (n,rexp n) :: (eqns (n-1))

            in rev (eqns newStart) 
            end

        and member x xs = List.exists (fn y => y=x) xs
    end;

structure RegEqnsFm : REGEQNSFM = RegEqnsFm(structure E = RegEqns structure M = Fm)



(* Example:

open RegEqnsFm;

val ndM = M.make{trans = [(0,"a",1),(0,"a",2),
                          (1,"a",0),(1,"b",2),
                          (2,"b",0),(2,"b",1)], 
                 start=[0,2],
                 final=[1,2]};

Compiler.Control.Print.printDepth := 10

fmToEqns ndM;

val it =
  [(0,Plus [Times [Atom "a",Var 1],Times [Atom "a",Var 2]]),
   (1,Plus [Times [],Times [Atom "a",Var 0],Times [Atom "b",Var 2]]),
   (2,Plus [Times [],Times [Atom "b",Var 0],Times [Atom "b",Var 1]]),
   (3,Plus [Var 0,Var 2])] : string re RegEqns.eqns
*)

(* Sollte zu ppregeqnsfm.sml:

val dM = M.determine ndM;display dM "test";       (* 4 states *)
M.show dM;
val it =
  {final=[2,3],
   trans=[(0,"b",1),(0,"a",2),
          (1,"b",1),(1,"a",1),
          (2,"b",3),(2,"a",0),
          (3,"b",3),(3,"a",3)],
   start=[0]}

display (retofm(fmtore ndM)) "test"   (* 91 states! *)
display ((M.minimize o M.determine) (retofm(fmtore dM))) "test"; (* 4 states *)
display (retofm(fmtore dM)) "test";   (* 21 states *)


Compiler.Control.Print.printDepth := 10;             
fmtore dM;   
- val it =
  Plus
    [Times [Star (Times [Atom "a",Atom "a"]),Atom "a"],
     Times
       [Star (Times [Atom "a",Atom "a"]),Atom "a",Atom "b",
        Star (Plus [Atom "a",Atom "b"])]] : string re
Frueher mit regSys war es nicht ausmultiplziert:
val it =
  Times
    [Atom "a",Star (Times [Atom "a",Atom "a"]),
     Plus [Times [],Times [Atom "b",Star (Plus [Atom "b",Atom "a"])]]]
(* = a(aa)*+a(aa)*b(b+a)* ok *)


Example 5 (Translation from regular expressions (a+b)^*a(a+b)^n to finite machines)

structure T = ppRegExpFm(StringLabels);

val a = T.E.Atom "a" and b = T.E.Atom "b"

fun exp n =  T.E.Times[T.E.Star(T.E.Plus[a,b]),a,T.E.Power (T.E.Plus[a,b]) n]
and mach n = T.M.minimize (T.M.determine (T.retofm (exp n)))
and disp n = T.display (mach n) "test"

T.M.states (mach 2); (*  8 *)
T.M.states (mach 3); (* 16 *)
T.M.states (mach 4); (* 32 *)
T.M.states (mach 5); (* 64 *)
T.M.states (mach 6); (* 128 *)

T.fmtore (mach 0); 
T.fmtore (mach 1); 
T.fmtore (mach 2); (* langsam! *)

disp 3;
disp 4;
*)
