(* File:    regexp.sml
   Content: - Regular expressions over alphabet 'a. 
            - Naming facility for variables and atoms.
            - Pretty printing of regular expressions and recursion equation systems 
              with named atomic expressions. (replaces former prVar)
   Author: H.Leiss, 2/99 

   Status:   uses SML-Site-Library/set.sml 
   - 19.2.9
     dnf: corrected  fun F [] prods = prod  to  fun F [] prods = []  
     maxVar: had been maxvar
     subst:  had been simSubst
     shiftVars: removed to language interpretation
   - 3.3.99
     renamed leftFactorize to factorizel, rightFactorize to factorizer
     added structure StandardNames
     PPREGEXP subsumes REGEXP_WEAK
   - 11.11.99 ppStar: use brackets (r)*; needed for (a;b)* etc.
   - 23.12.99 moved some signatures to regexp.new.sig
   - 27.12.99 removed rdnf, rightFactors and factorizeRight
   - 10.2.2000 moved structure StandardNames to Site-Library/names.sml
   - 4.3.2000 removed boundVars

CM.set_root "regeqnsNFnew.cm";

Moved to regexp.new.sig:          Moved to label.sig:     Moved to names.sml
                  
signature REGEXP_WEAK             signature NAMES         structure StandardNames
signature PPREGEXP
*)


signature REGEXP =
  sig
    datatype 'a re
      = Atom of 'a
      | Plus of 'a re list
      | Star of 'a re
      | Times of 'a re list
      | Var of int

    val Unit   : 'a re
    val Zero   : 'a re
    val It     : 'a re -> 'a re
    val Power  : 'a re -> int -> 'a re
    val Powers : 'a re -> int -> 'a re

    exception Undefined

    val freeVars : 'a re -> int Set.t
    val subst : (int * 'a re) list -> 'a re -> 'a re
    val substExp : ''a re * ''a re -> ''a re -> ''a re
    val varNum : 'a re -> int
    val maxVar : 'a re -> int

    val nf : ''a re -> ''a re
    val dnf : ''a re -> ''a re

    val factorizel : ''a re -> ''a re
    val factorizer : ''a re -> ''a re
  end

(* ----------  Regular Expressions over alphabet 'a  ----------- *)

structure RegExp : REGEXP = 
    struct
	datatype 'a re = 
	    Var of int
	  | Atom of 'a 
	  | Star of ('a re)
	  | Plus of ('a re) list    (* Sum,Prod are already used in SML'90 *)
	  | Times of ('a re) list

	val Zero = Plus []          (* Empty(-) : exn  exists in SML'97 *)
	val Unit = Times []
	fun It r = Times [r, Star r]
	fun Power r n = 
	    let 
		fun factors n = if n <= 0 then [] else r::(factors (n-1))
	    in 
		Times (factors n)
	    end
	and Powers r n = 
	    let 
		fun powers n =
		    if n <= 0 then [Times []] else (Power r n) :: (powers (n-1)) 
	    in 
		Plus (powers n)
	    end

	fun freeVars (Var i) = Set.mk [i]
	  | freeVars (Atom a) = Set.mk []
	  | freeVars (Star r) = freeVars r
	  | freeVars (Plus rs)  = Set.Union (map freeVars rs)
	  | freeVars (Times rs) = Set.Union (map freeVars rs)

	fun maxVar (Var i) = i
	  | maxVar (Atom a) = ~1
	  | maxVar (Star r) = maxVar r
	  | maxVar (Plus rs)  = foldl Int.max ~1 (map maxVar rs)
	  | maxVar (Times rs) = foldl Int.max ~1 (map maxVar rs)

	exception Undefined

	fun varNum (Var i) = i 
	  | varNum _ = raise Undefined

	fun subst eqns (Atom a) = Atom a
	  | subst eqns (Var j)  = 
	    (case (List.find (fn (i,r) => i=j) eqns) of
		 SOME (i,r) => r | NONE => Var j)
	  | subst eqns (Plus rs) = Plus (map (subst eqns) rs)
	  | subst eqns (Times rs)= Times (map (subst eqns) rs)
	  | subst eqns (Star r) = Star (subst eqns r)

	(* Substitute expression t for s in r *)

	fun substExp (s,t) r =  
	    if (s = r) then t else 
		(case r of 
		     (Atom a) => (Atom a) 
		   | (Var j)  => (Var j)
		   | (Plus rs)  => Plus (map (substExp (s,t)) rs)
		   | (Times rs) => Times (map (substExp (s,t)) rs)
		   | (Star r)   => Star (substExp (s,t) r))

	(* Quasi normal form: we use

	 - associativity of Plus:  (x+(y+z)) => (x+y+z) <= ((x+y)+z)
	 - associativity of Times: (x;(y;z)) => (x;y;z) <= ((x;y);z)
	 - idempotency of Plus:      (x + x) => x
         - neutrality of Zero and Unit:  0+x => x <= 1;x 
         - annihilation by Zero:         0;x => 0 <= x;0
         - transitivity of Star:         x** => x*
         - reflexivity of Star:     [x + 1]* => x*
         - Star on constants:             1* => 1 <= 0*
         - Star on sums:           [x + y*]* => (x + y)*               *)

	fun nf (Var j) = Var j
	  | nf (Atom a) = Atom a
	  | nf (Plus []) = Zero
	  | nf (Plus (r::rs)) = 
	    let 
		val (nfr,nfrs) = (nf r, nf (Plus rs))
		fun union [] U = U
		  | union (q::Q) U = union Q (add q U)
		and add q [] = [q]
		  | add q (u::U) = if q=u then (u::U) else (u::(add q U))
	    in
		(case (nfr, nfrs) of
		     (Plus [],_) => nfrs
		   | (_,Plus []) => nfr
		   | (Plus R, Plus RS) => (Plus (union RS R))
		   | (_,Plus RS) => Plus (union RS [nfr])
		   | (Plus R,_) => Plus (union R [nfrs])
		   | _ => if nfr=nfrs then nfr else Plus [nfr,nfrs])
	    end
	  | nf (Times []) = Unit
	  | nf (Times (r::rs)) = 
	    let
		val (nfr, nfrs) = (nf r, nf (Times rs))
	    in 
		case (nfr, nfrs) of
		    (Times [],_) => nfrs
		  | (_,Times []) => nfr
		  | (Plus [],_) => Zero
		  | (_,Plus []) => Zero
		  | (Times R,Times Rs) => Times (R @ Rs)
		  | (_,Times Rs) => Times (nfr :: Rs)
		  | (Times R,_) => Times (R @ [nfrs])
		  | _ => Times [nfr, nfrs]
	    end
	  | nf (Star r) = 
	    let 
		val nfr = nf r 
	    in 
		case nfr of 
		    Plus [] => Unit
		  | Plus ts => 
			(case (List.filter (fn (Times[]) => false | _ => true)
					  (map (fn (Star t) => t | t => t) ts)) 
			     of 
				 [] => Unit 
			       | [s] => Star s
			       | ss => Star (Plus ss))
		  | Times [] => Unit 
		  | (Star R) => nfr
		  | _ => (Star nfr)
	    end

	(* Disjunctive normal form of a regular expression; does not change (Star r) *)

	fun dnf (Atom a) = Atom a
	  | dnf (Var i) = Var i
	  | dnf (Plus rs) = nf (Plus (map dnf rs))
	  | dnf (Times []) = Unit
	  | dnf (Times (s::ss)) =
	    let
		val (dnfs,dnfss) = (dnf s, dnf (Times ss))
	    in 
		case (dnfs,dnfss) of 
		    (Plus ps, Plus pps) => 
			let 
			    val fns = map (fn x => fn y => Times [x,y]) ps
			    fun F [] prods = []
			      | F (f::fs) prods = (map f prods) @ (F fs prods)
			in
			    nf (Plus (F fns pps))
			end
		  | (Plus ps, _) => nf (Plus (map (fn y => Times [y,dnfss]) ps))
		  | (_, Plus pps) => nf (Plus (map (fn y => Times [dnfs,y]) pps))
		  | (_, _) => nf (Times [dnfs, dnfss])
	    end
	  | dnf (Star r) = (Star r) 


	(* Extract common atomic left factors: 
	 
	 In a sum of products of atomic expressions, collect all summands of form 
	 Times [Var i, r1,...,rk] and extract their common left factor (Var i).
	 Similarly extract common left factors (Atom a).
	 *)

	fun factorizel e =
	    let 
		fun leftFact (Atom a) acc = leftFact (Plus [Atom a]) acc
		  | leftFact (Var i) acc = leftFact (Plus [Var i]) acc
		  | leftFact (Times ss) acc = leftFact (Plus [(Times ss)]) acc
		  | leftFact (Star r) acc = leftFact (Plus [Star r]) acc
		  | leftFact (Plus ((Times [s,t])::rs)) acc =
		    leftFact (Plus rs) (add (s,t) acc)
		  | leftFact (Plus ((Times (s::ss))::rs)) acc =
		    leftFact (Plus rs) (add (s,Times ss) acc)
		  | leftFact (Plus (r::rs)) acc = 
		    leftFact (Plus rs) (add (r,Unit) acc)
		  | leftFact (Plus []) acc = 
		    (Plus (map (fn (r,rs) => case rs of 
				[s] => Times [r,s] 
			      | _ => Times[r,Plus rs]) (rev acc)))

		and add (r,s) [] = [(r,[s])]
		  | add (r,s) ((t,ts)::rs) = 
		    if r=t then ((r,s::ts)::rs) else ((t,ts)::(add (r,s) rs))
	    in 
		leftFact e []
	    end

	fun factorizer e =
	    let 
		fun rightFact (Plus []) acc = 
		    (Plus (map (fn (r,rs) => 
				case rs of [s] => Times [s,r] | _ => Times[Plus rs,r])
			   (rev acc)))

		  | rightFact (Plus ((Times es)::rs)) acc =
		    (case (rev es) of 
			 []      => rightFact (Plus rs) (add (Times[],Times[]) acc)
		       | [s,t]   => rightFact (Plus rs) (add (s,t) acc)
		       | (s::ss) => rightFact (Plus rs) (add (s,Times (rev ss)) acc))
		  | rightFact (Plus (r::rs)) acc = rightFact (Plus rs) (add (r,Unit) acc)
		  | rightFact (Times ss) acc = rightFact (Plus [Times ss]) acc
		  | rightFact (Star s) acc = rightFact (Plus [Star s]) acc
		  | rightFact (Atom a) acc = rightFact (Plus [Atom a]) acc
		  | rightFact (Var i) acc = rightFact (Plus [Var i]) acc

		and add (r,s) [] = [(r,[s])]
		  | add (r,s) ((t,ts)::rs) = 
		    if r=t then ((r,s::ts)::rs) else ((t,ts)::(add (r,s) rs))
	    in 
		rightFact e []
	    end
    end		



